from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.builtin.interop.interop import Interop
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestEvent(BoaTest):
    default_folder: str = 'test_sc/event_test'

    def test_event_without_arguments(self):
        event_id = 'Event'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.NEWARRAY0    # Main()
            + Opcode.PUSHDATA1      # event()
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET       # return
        )

        path = self.get_contract_path('EventWithoutArguments.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((), event_notifications[0].arguments)

    def test_event_with_arguments(self):
        event_id = 'Event'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.PUSH10        # Main()
            + Opcode.PUSH1          # event(10)
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET       # return
        )

        path = self.get_contract_path('EventWithArgument.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((10,), event_notifications[0].arguments)

    def test_event_with_name(self):
        event_id = 'example'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.PUSH10       # Main()
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1      # event(10)
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET      # return
        )

        path = self.get_contract_path('EventWithName.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((10,), event_notifications[0].arguments)

    def test_event_with_annotation(self):
        event_id = 'example'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.PUSH10       # Main()
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.PUSHDATA1      # on_example(10)
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET      # return
        )

        path = self.get_contract_path('EventWithAnnotation.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((10,), event_notifications[0].arguments)

    def test_event_nep5_transfer(self):
        event_id = 'transfer'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.INITSLOT     # Main()
            + b'\x00\x03'
            + Opcode.LDARG2         # event(from_addr, to_addr, amount)
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET       # return
        )

        path = self.get_contract_path('EventNep5Transfer.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'1', b'2', 10)
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual(('1', '2', 10), event_notifications[0].arguments)

    def test_event_nep17_transfer(self):
        event_id = 'Transfer'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.INITSLOT     # Main()
            + b'\x00\x03'
            + Opcode.LDARG2         # event(from_addr, to_addr, amount)
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET       # return
        )

        path = self.get_contract_path('EventNep17Transfer.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'1', b'2', 10)
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual(('1', '2', 10), event_notifications[0].arguments)

    def test_event_nep17_transfer_built(self):
        event_id = 'Transfer'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.INITSLOT     # Main()
            + b'\x00\x03'
            + Opcode.LDARG2         # event(from_addr, to_addr, amount)
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET       # return
        )

        path = self.get_contract_path('EventNep17TransferBuilt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'1', b'2', 10)
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual(('1', '2', 10), event_notifications[0].arguments)

    def test_event_nep11_transfer(self):
        event_id = 'Transfer'
        event_name = String(event_id).to_bytes(min_length=1)
        expected_output = (
            Opcode.INITSLOT     # Main()
            + b'\x00\x04'
            + Opcode.LDARG3         # event(from_addr, to_addr, amount, token_id)
            + Opcode.LDARG2
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.PUSH4
            + Opcode.PACK
            + Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET       # return
        )

        path = self.get_contract_path('EventNep11Transfer.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', b'1', b'2', 10, b'someToken')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual(('1', '2', 10, 'someToken'), event_notifications[0].arguments)

    def test_event_with_return(self):
        path = self.get_contract_path('EventWithoutTypes.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_event_with_duplicated_name(self):
        path = self.get_contract_path('EventWithDuplicatedName.py')

        engine = TestEngine()
        arg = 10
        event_id = 'example'
        result = self.run_smart_contract(engine, path, 'example', arg)
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=event_id)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((arg,), event_notifications[0].arguments)

    def test_event_call_mismatched_type(self):
        path = self.get_contract_path('MismatchedTypeCallEvent.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_event_with_abort(self):
        path = self.get_contract_path('EventWithAbort.py')
        self.compile_and_save(path)

        engine = TestEngine()

        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, path, 'send_event_with_abort')
        self.assertEqual(0, len(engine.notifications))

        result = self.run_smart_contract(engine, path, 'send_event')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        previous_notification_count = len(engine.notifications)
        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            """
            Notifications on a fault execution aren't logged into the blockchain
            Even with rollback_on_fault set as False, TestEngine doesn't return it because that's the expected behavior
            """
            self.run_smart_contract(engine, path, 'send_event_with_abort', rollback_on_fault=False)

        self.assertEqual(previous_notification_count, len(engine.notifications))

    def test_boa2_event_test(self):
        path = self.get_contract_path('EventBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(7, result)

        event_notifications1 = engine.get_events('transfer_test')
        self.assertEqual(1, len(event_notifications1))
        self.assertEqual(3, len(event_notifications1[0].arguments))
        from_address, to_address, amount = event_notifications1[0].arguments
        self.assertEqual(2, from_address)
        self.assertEqual(5, to_address)
        self.assertEqual(7, amount)

        event_notifications2 = engine.get_events('refund')
        self.assertEqual(1, len(event_notifications2))
        self.assertEqual(2, len(event_notifications2[0].arguments))
        to_address, amount = event_notifications2[0].arguments
        self.assertEqual('me', to_address)
        self.assertEqual(52, amount)



from __future__ import annotations

from typing import List

import boa3
from boa3.neo.contracts import NEF, Version
from boa3.neo.core import BinaryReader, BinaryWriter
from boa3.neo3.contracts.nef import MethodToken


class NefFile:
    """
    The object encapsulates the information of the NEO Executable Format (NEF)

    :ivar _nef: nef serializable object
    """

    def __init__(self, script_bytes: bytes,
                 method_tokens: List[MethodToken] = None,
                 source: str = None):
        """
        :param script_bytes: the script of the smart contract
        """
        compiler: str = f"neo3-boa by COZ-{Version.from_string(boa3.__version__)}"
        self._nef = NEF(compiler_name=compiler,
                        script=script_bytes,
                        tokens=[] if method_tokens is None else method_tokens,
                        source=source)

    @property
    def script(self) -> bytes:
        return self._nef.script

    @property
    def script_hash(self) -> bytes:
        from boa3.neo.cryptography import hash160
        return hash160(self.script)

    @property
    def source(self) -> str:
        return self._nef.source

    @property
    def checksum(self) -> int:
        return self._nef.checksum

    def serialize(self) -> bytes:
        """
        Serialize the NefFile object

        :return: bytes of the serialized object.
        """
        with BinaryWriter() as writer:
            self._nef.serialize(writer)
            result = writer.to_array()
        return result

    @classmethod
    def deserialize(cls, bts: bytes) -> NefFile:
        """
        Deserialize the NefFile object

        :param bts: stream data
        :return: the deserialized object.
        :rtype: NefFile
        """
        with BinaryReader(bts) as reader:
            nef = NEF()
            nef.deserialize(reader)

            nef_file = cls(nef.script)
            nef_file._nef = nef
        return nef_file



# -------------------------------------------
# This standard is deprecated
# Please check the NEP17 example
# -------------------------------------------

from typing import Any

from boa3.builtin.compile_time import NeoMetadata, metadata, public
from boa3.builtin.contract import Nep5TransferEvent
from boa3.builtin.interop import runtime, storage
from boa3.builtin.type import UInt160


# -------------------------------------------
# METADATA
# -------------------------------------------


@metadata
def manifest_metadata() -> NeoMetadata:
    """
    Defines this smart contract's metadata information
    """
    meta = NeoMetadata()
    meta.supported_standards = ['NEP-5']

    meta.author = "Mirella Medeiros and Ricardo Prado. COZ in partnership with Simpli"
    meta.description = "NEP-5 Example"
    meta.email = "contact@coz.io"
    return meta


# -------------------------------------------
# TOKEN SETTINGS
# -------------------------------------------


# Script hash of the contract owner
OWNER = UInt160(b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00')
SUPPLY_KEY = 'totalSupply'

# Name of the Token
TOKEN_NAME = 'NEP5 Standard'

# Symbol of the Token
TOKEN_SYMBOL = 'NEP5'

# Number of decimal places
TOKEN_DECIMALS = 8

# Total Supply of tokens in the system
TOKEN_TOTAL_SUPPLY = 10_000_000 * 10 ** TOKEN_DECIMALS  # 10m total supply * 10^8 (decimals)

# -------------------------------------------
# Events
# -------------------------------------------


on_transfer = Nep5TransferEvent


# -------------------------------------------
# Methods
# -------------------------------------------


@public
def verify() -> bool:
    """
    When this contract address is included in the transaction signature,
    this method will be triggered as a VerificationTrigger to verify that the signature is correct.
    For example, this method needs to be called when withdrawing token from the contract.

    :return: whether the transaction signature is correct
    """
    return runtime.check_witness(OWNER)


@public
def name() -> str:
    """
    Gets the name of the token.

    This method must always return the same value every time it is invoked.

    :return: the name of the token.
    """
    return TOKEN_NAME


@public
def symbol() -> str:
    """
    Gets the symbols of the token.

    This symbol should be short (3-8 characters is recommended), with no whitespace characters or new-lines and should
    be limited to the uppercase latin alphabet (i.e. the 26 letters used in English).
    This method must always return the same value every time it is invoked.

    :return: a short string symbol of the token managed in this contract.
    """
    return TOKEN_SYMBOL


@public
def decimals() -> int:
    """
    Gets the amount of decimals used by the token.

    E.g. 8, means to divide the token amount by 100,000,000 (10 ^ 8) to get its user representation.
    This method must always return the same value every time it is invoked.

    :return: the number of decimals used by the token.
    """
    return TOKEN_DECIMALS


@public
def totalSupply() -> int:
    """
    Gets the total token supply deployed in the system.

    This number mustn't be in its user representation. E.g. if the total supply is 10,000,000 tokens, this method
    must return 10,000,000 * 10 ^ decimals.

    :return: the total token supply deployed in the system.
    """
    return TOKEN_TOTAL_SUPPLY


@public
def balanceOf(account: bytes) -> int:
    """
    Get the current balance of an address

    The parameter account should be a 20-byte address.

    :param account: the account address to retrieve the balance for
    :type account: bytes

    :return: the token balance of the `account`
    :raise AssertionError: raised if `account` length is not 20.
    """
    assert len(account) == 20
    return storage.get(account).to_int()


@public
def transfer(from_address: UInt160, to_address: UInt160, amount: int) -> bool:
    """
    Transfers a specified amount of NEP5 tokens from one account to another

    If the method succeeds, it must fire the `transfer` event and must return true, even if the amount is 0,
    or from and to are the same address.

    :param from_address: the address to transfer from
    :type from_address: UInt160
    :param to_address: the address to transfer to
    :type to_address: UInt160
    :param amount: the amount of NEP5 tokens to transfer
    :type amount: int

    :return: whether the transfer was successful
    :raise AssertionError: raised if `from_address` or `to_address` length is not 20 or if `amount` is less than zero.
    """
    # the parameters from and to should be 20-byte addresses. If not, this method should throw an exception.
    assert len(from_address) == 20 and len(to_address) == 20
    # the parameter amount must be greater than or equal to 0. If not, this method should throw an exception.
    assert amount >= 0

    # The function MUST return false if the from account balance does not have enough tokens to spend.
    from_balance = storage.get(from_address).to_int()
    if from_balance < amount:
        return False

    # The function should check whether the from address equals the caller contract hash.
    # If so, the transfer should be processed;
    # If not, the function should use the check_witness to verify the transfer.
    if from_address != runtime.calling_script_hash:
        if not runtime.check_witness(from_address):
            return False

    # if the `to_address` is a deployed contract, the function should check the payable flag of this contract
    # TODO: include example when objects are implemented

    if from_address == to_address:
        # transfer to self
        return True

    if from_balance == amount:
        storage.delete(from_address)
    else:
        storage.put(from_address, from_balance - amount)

    to_balance = storage.get(to_address).to_int()
    storage.put(to_address, to_balance + amount)

    # if the method succeeds, it must fire the transfer event, and must return true
    on_transfer(from_address, to_address, amount)
    return True


@public
def _deploy(data: Any, update: bool):
    """
    Initializes the storage when the smart contract is deployed.

    :return: whether the deploy was successful. This method must return True only during the smart contract's deploy.
    """
    if not update:
        storage.put(SUPPLY_KEY, TOKEN_TOTAL_SUPPLY)
        storage.put(OWNER, TOKEN_TOTAL_SUPPLY)

        on_transfer(b'', OWNER, TOKEN_TOTAL_SUPPLY)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestFor(BoaTest):
    default_folder: str = 'test_sc/for_test'

    def test_for_tuple_condition(self):
        jmpif_address = Integer(19).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-22).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH15     # for_sequence = (3, 5, 15)
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP        # begin for
            + jmpif_address
            + Opcode.OVER           # x = for_sequence[for_index]
            + Opcode.OVER
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC1
            + Opcode.LDLOC0         # a = a + x
            + Opcode.LDLOC1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.INC            # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end for
            + jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('TupleCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(23, result)

    def test_for_string_condition(self):
        path = self.get_contract_path('StringCondition.py')
        output = Boa3.compile(path)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual('5', result)

    def test_for_mismatched_type_condition(self):
        path = self.get_contract_path('MismatchedTypeCondition.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_for_no_condition(self):
        path = self.get_contract_path('NoCondition.py')

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_nested_for(self):
        outer_jmpif_address = Integer(47).to_byte_array(min_length=1, signed=True)
        outer_jmp_address = Integer(-50).to_byte_array(min_length=1, signed=True)

        inner_jmpif_address = Integer(21).to_byte_array(min_length=1, signed=True)
        inner_jmp_address = Integer(-24).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x04'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH15     # sequence = (3, 5, 15)
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.LDLOC1     # outer_for_sequence = sequence
            + Opcode.PUSH0      # outer_for_index = 0
            + Opcode.JMP
            + outer_jmpif_address
            + Opcode.OVER           # x = outer_for_sequence[outer_for_index]
            + Opcode.OVER
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC2
            + Opcode.LDLOC1     # inner_for_sequence = sequence
            + Opcode.PUSH0      # inner_for_index = 0
            + Opcode.JMP
            + inner_jmpif_address
            + Opcode.OVER         # y = inner_for_sequence[inner_for_index]
            + Opcode.OVER
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC3
            + Opcode.LDLOC0         # a = a + x * y
            + Opcode.LDLOC2
            + Opcode.LDLOC3
            + Opcode.MUL
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.INC            # inner_for_index = inner_for_index + 1
            + Opcode.DUP        # if inner_for_index < len(inner_for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end inner_for
            + inner_jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.INC     # outer_for_index = outer_for_index + 1
            + Opcode.DUP        # if outer_for_index < len(outer_for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end outer_for
            + outer_jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('NestedFor.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(529, result)

    def test_for_else(self):
        jmpif_address = Integer(19).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-22).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH15     # sequence = (3, 5, 15)
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.LDLOC1     # for_sequence = sequence
            + Opcode.PUSH0      # for_index = 0
            + Opcode.JMP
            + jmpif_address
            + Opcode.OVER           # x = for_sequence[for_index]
            + Opcode.OVER
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.STLOC2
            + Opcode.LDLOC0         # a = a + x
            + Opcode.LDLOC2
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.INC            # for_index = for_index + 1
            + Opcode.DUP        # if for_index < len(for_sequence)
            + Opcode.PUSH2
            + Opcode.PICK
            + Opcode.SIZE
            + Opcode.LT
            + Opcode.JMPIF      # end for
            + jmp_address
            + Opcode.DROP
            + Opcode.DROP
            + Opcode.LDLOC0     # a = a + 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('ForElse.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(24, result)

    def test_for_continue(self):
        path = self.get_contract_path('ForContinue.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(20, result)

    def test_for_break(self):
        path = self.get_contract_path('ForBreak.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(6, result)

    def test_for_break_else(self):
        path = self.get_contract_path('ForBreakElse.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(6, result)

    def test_for_iterate_dict(self):
        path = self.get_contract_path('ForIterateDict.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_boa2_iteration_test(self):
        path = self.get_contract_path('IterBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(18, result)

    def test_boa2_iteration_test2(self):
        path = self.get_contract_path('IterBoa2Test2.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(8, result)

    def test_boa2_iteration_test3(self):
        path = self.get_contract_path('IterBoa2Test3.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(7, result)

    def test_boa2_iteration_test4(self):
        path = self.get_contract_path('IterBoa2Test4.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_boa2_iteration_test5(self):
        path = self.get_contract_path('IterBoa2Test5.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(51, result)

    def test_boa2_iteration_test6(self):
        path = self.get_contract_path('IterBoa2Test6.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(10, result)

    def test_boa2_iteration_test7(self):
        path = self.get_contract_path('IterBoa2Test7.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(12, result)

    def test_boa2_iteration_test8(self):
        path = self.get_contract_path('IterBoa2Test8.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(6, result)

    def test_for_pass(self):
        path = self.get_contract_path('ForPass.py')
        engine = TestEngine()

        output = Boa3.compile(path)
        self.assertIn(Opcode.NOP, output)

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsVoid(result)

    def test_for_range(self):
        path = self.get_contract_path('ForWithContractInterface.py')
        path_contract_called = self.get_contract_path('ForWithContractInterfaceCalled.py')
        engine = TestEngine()
        self.run_smart_contract(engine, path_contract_called, 'return_zero')

        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual([0, 0, 0], result)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestTuple(BoaTest):
    default_folder: str = 'test_sc/tuple_test'

    def test_tuple_int_values(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH3      # a = (1, 2, 3)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('IntTuple.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_string_values(self):
        byte_input0 = String('1').to_bytes()
        byte_input1 = String('2').to_bytes()
        byte_input2 = String('3').to_bytes()

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = ('1', '2', '3')
            + Integer(len(byte_input2)).to_byte_array()
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array()
            + byte_input1
            + Opcode.PUSHDATA1
            + Integer(len(byte_input0)).to_byte_array()
            + byte_input0
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('StrTuple.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_bool_values(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHF      # a = (True, True, False)
            + Opcode.PUSHT
            + Opcode.PUSHT
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('BoolTuple.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_variable_values(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x04'
            + b'\x00'
            + Opcode.PUSH1      # a = 1
            + Opcode.STLOC0
            + Opcode.PUSH2      # b = 2
            + Opcode.STLOC1
            + Opcode.PUSH3      # c = 3
            + Opcode.STLOC2
            + Opcode.PUSH3      # d = (a, b, c)
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3      # tuple length
            + Opcode.PACK
            + Opcode.STLOC3
            + Opcode.RET        # return
        )

        path = self.get_contract_path('VariableTuple.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_assign_empty_tuple(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.NEWARRAY0
            + Opcode.STLOC0
            + Opcode.RET        # return
        )

        path = self.get_contract_path('EmptyTupleAssignment.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_tuple_get_value(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('GetValue.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', [1, 2, 3, 4])
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', [5, 3, 2])
        self.assertEqual(5, result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'Main', [])

    def test_non_sequence_get_value(self):
        path = self.get_contract_path('MismatchedTypeGetValue.py')
        self.assertCompilerLogs(CompilerError.UnresolvedOperation, path)

    def test_tuple_set_value(self):
        path = self.get_contract_path('SetValue.py')
        self.assertCompilerLogs(CompilerError.UnresolvedOperation, path)

    def test_non_sequence_set_value(self):
        path = self.get_contract_path('MismatchedTypeSetValue.py')
        self.assertCompilerLogs(CompilerError.UnresolvedOperation, path)

    def test_tuple_index_mismatched_type(self):
        path = self.get_contract_path('MismatchedTypeTupleIndex.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_tuple_of_tuple(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x01'
            + Opcode.LDARG0     # arg[0][0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('TupleOfTuple.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', ((1, 2), (3, 4)))
        self.assertEqual(result, 1)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'Main', ())

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'Main', ((), (1, 2), (3, 4)))

    def test_nep5_main(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG1     # args[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET        # return
        )

        path = self.get_contract_path('Nep5Main.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'op', (1, 2, 3, 4))
        self.assertEqual(1, result)
        result = self.run_smart_contract(engine, path, 'Main', 'op', ('a', False))
        self.assertEqual('a', result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_IS_OUT_OF_RANGE_MSG_REGEX_SUFFIX):
            self.run_smart_contract(engine, path, 'Main', 'op', ())

    def test_tuple_slicing(self):
        path = self.get_contract_path('TupleSlicingLiteralValues.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2], result)

    def test_tuple_slicing_start_larger_than_ending(self):
        path = self.get_contract_path('TupleSlicingStartLargerThanEnding.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([], result)

    def test_tuple_slicing_with_variables(self):
        path = self.get_contract_path('TupleSlicingVariableValues.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2], result)

    def test_tuple_slicing_negative_start(self):
        path = self.get_contract_path('TupleSlicingNegativeStart.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2, 3, 4, 5], result)

    def test_tuple_slicing_negative_end(self):
        path = self.get_contract_path('TupleSlicingNegativeEnd.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1], result)

    def test_tuple_slicing_start_omitted(self):
        path = self.get_contract_path('TupleSlicingStartOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1, 2], result)

    def test_tuple_slicing_omitted(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH5      # a = (0, 1, 2, 3, 4, 5)
            + Opcode.PUSH4
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH0
            + Opcode.PUSH6
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a[:]
            + Opcode.UNPACK
            + Opcode.PACK
            + Opcode.RET        # return
        )
        path = self.get_contract_path('TupleSlicingOmitted.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([0, 1, 2, 3, 4, 5], result)

    def test_tuple_slicing_end_omitted(self):
        path = self.get_contract_path('TupleSlicingEndOmitted.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual([2, 3, 4, 5], result)

    def test_tuple_slicing_with_stride(self):
        path = self.get_contract_path('TupleSlicingWithStride.py')
        engine = TestEngine()

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[2:5:2]
        result = self.run_smart_contract(engine, path, 'literal_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-6:5:2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[0:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-6:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-999:5:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[0:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-999:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[999:5:2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[0:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[999:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_values')
        self.assertEqual(expected_result, tuple(result))

    def test_tuple_slicing_with_negative_stride(self):
        path = self.get_contract_path('TupleSlicingWithNegativeStride.py')
        engine = TestEngine()

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[2:5:-1]
        result = self.run_smart_contract(engine, path, 'literal_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-6:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[0:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-6:-1:-1]
        result = self.run_smart_contract(engine, path, 'negative_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-999:5:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[0:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-999:-999:-1]
        result = self.run_smart_contract(engine, path, 'negative_really_low_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[999:5:-1]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[0:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[999:999:-1]
        result = self.run_smart_contract(engine, path, 'really_high_values')
        self.assertEqual(expected_result, tuple(result))

    def test_tuple_slicing_omitted_with_stride(self):
        path = self.get_contract_path('TupleSlicingOmittedWithStride.py')
        engine = TestEngine()

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[::2]
        result = self.run_smart_contract(engine, path, 'omitted_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:5:2]
        result = self.run_smart_contract(engine, path, 'omitted_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[2::2]
        result = self.run_smart_contract(engine, path, 'omitted_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-6::2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:-1:2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-999::2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:-999:2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[999::2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:999:2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, tuple(result))

    def test_tuple_slicing_omitted_with_negative_stride(self):
        path = self.get_contract_path('TupleSlicingOmittedWithNegativeStride.py')
        engine = TestEngine()

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[::-2]
        result = self.run_smart_contract(engine, path, 'omitted_values')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:5:-2]
        result = self.run_smart_contract(engine, path, 'omitted_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[2::-2]
        result = self.run_smart_contract(engine, path, 'omitted_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-6::-2]
        result = self.run_smart_contract(engine, path, 'negative_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:-1:-2]
        result = self.run_smart_contract(engine, path, 'negative_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[-999::-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:-999:-2]
        result = self.run_smart_contract(engine, path, 'negative_really_low_end')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[999::-2]
        result = self.run_smart_contract(engine, path, 'really_high_start')
        self.assertEqual(expected_result, tuple(result))

        a = (0, 1, 2, 3, 4, 5)
        expected_result = a[:999:-2]
        result = self.run_smart_contract(engine, path, 'really_high_end')
        self.assertEqual(expected_result, tuple(result))

    def test_tuple_index(self):
        path = self.get_contract_path('IndexTuple.py')
        engine = TestEngine()

        tuple_ = (1, 2, 3, 4)
        value = 3
        start = 0
        end = 4
        result = self.run_smart_contract(engine, path, 'main', tuple_, value, start, end)
        self.assertEqual(tuple_.index(value, start, end), result)

        tuple_ = (1, 2, 3, 4)
        value = 3
        start = 2
        end = 4
        result = self.run_smart_contract(engine, path, 'main', tuple_, value, start, end)
        self.assertEqual(tuple_.index(value, start, end), result)

        from boa3.model.builtin.builtin import Builtin
        with self.assertRaisesRegex(TestExecutionException, f'{Builtin.SequenceIndex.exception_message}$'):
            self.run_smart_contract(engine, path, 'main', (1, 2, 3, 4), 3, 3, 4)

        with self.assertRaisesRegex(TestExecutionException, f'{Builtin.SequenceIndex.exception_message}$'):
            self.run_smart_contract(engine, path, 'main', (1, 2, 3, 4), 3, 4, -1)

        with self.assertRaisesRegex(TestExecutionException, f'{Builtin.SequenceIndex.exception_message}$'):
            self.run_smart_contract(engine, path, 'main', (1, 2, 3, 4), 3, 0, -99)

        tuple_ = (1, 2, 3, 4)
        value = 3
        start = 0
        end = -1
        result = self.run_smart_contract(engine, path, 'main', tuple_, value, start, end)
        self.assertEqual(tuple_.index(value, start, end), result)

        tuple_ = (1, 2, 3, 4)
        value = 2
        start = 0
        end = 99
        result = self.run_smart_contract(engine, path, 'main', tuple_, value, start, end)
        self.assertEqual(tuple_.index(value, start, end), result)

    def test_tuple_index_end_default(self):
        path = self.get_contract_path('IndexTupleEndDefault.py')
        engine = TestEngine()

        tuple_ = (1, 2, 3, 4)
        value = 3
        start = 0
        result = self.run_smart_contract(engine, path, 'main', tuple_, value, start)
        self.assertEqual(tuple_.index(value, start), result)

        from boa3.model.builtin.builtin import Builtin
        with self.assertRaisesRegex(TestExecutionException, f'{Builtin.SequenceIndex.exception_message}$'):
            self.run_smart_contract(engine, path, 'main', (1, 2, 3, 4), 2, 99)

        with self.assertRaisesRegex(TestExecutionException, f'{Builtin.SequenceIndex.exception_message}$'):
            self.run_smart_contract(engine, path, 'main', (1, 2, 3, 4), 4, -1)

        tuple_ = (1, 2, 3, 4)
        value = 2
        start = -10
        result = self.run_smart_contract(engine, path, 'main', tuple_, value, start)
        self.assertEqual(tuple_.index(value, start), result)

    def test_tuple_index_defaults(self):
        path = self.get_contract_path('IndexTupleDefaults.py')
        engine = TestEngine()

        tuple_ = (1, 2, 3, 4)
        value = 3
        result = self.run_smart_contract(engine, path, 'main', tuple_, value)
        self.assertEqual(tuple_.index(value), result)

        tuple_ = (1, 2, 3, 4)
        value = 1
        result = self.run_smart_contract(engine, path, 'main', tuple_, value)
        self.assertEqual(tuple_.index(value), result)



__all__ = ['ContractManagementClass',
           'CryptoLibClass',
           'GasClass',
           'LedgerClass',
           'NeoClass',
           'PolicyClass',
           'RoleManagementClass',
           'StdLibClass',
           ]

from boa3.model.builtin.native.contractmanagementclass import ContractManagementClass
from boa3.model.builtin.native.cryptolibclass import CryptoLibClass
from boa3.model.builtin.native.gasclass import GasClass
from boa3.model.builtin.native.ledgerclass import LedgerClass
from boa3.model.builtin.native.neoclass import NeoClass
from boa3.model.builtin.native.policyclass import PolicyClass
from boa3.model.builtin.native.rolemanagementclass import RoleManagementClass
from boa3.model.builtin.native.stdlibclass import StdLibClass



import ast
from typing import Dict, List, Tuple

from boa3.model.builtin.method.strmethod import StrMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class StrByteStringMethod(StrMethod):

    def __init__(self):
        from boa3.model.type.primitive.bytestringtype import ByteStringType
        from boa3.model.type.type import Type
        args: Dict[str, Variable] = {
            'object': Variable(ByteStringType.build()),
        }
        object_default = ast.parse("'{0}'".format(Type.str.default_value)).body[0].value

        super().__init__(args, [object_default])

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []



import ast
from typing import Dict

from boa3.model.builtin.interop.nativecontract import StdLibMethod
from boa3.model.variable import Variable


class ItoaMethod(StdLibMethod):

    def __init__(self, internal_call_args: int = None):
        from boa3.model.type.type import Type
        identifier = 'itoa'
        syscall = 'itoa'
        args: Dict[str, Variable] = {
            'value': Variable(Type.int),
            'base': Variable(Type.int)
        }
        args_default = ast.parse("{0}".format(10)).body[0].value

        super().__init__(identifier, syscall, args, defaults=[args_default], return_type=Type.str,
                         internal_call_args=internal_call_args)



from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.builtin.interop.interop import Interop
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3.neo3.contracts import CallFlags
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestRoleManagementClass(BoaTest):
    default_folder: str = 'test_sc/native_test/rolemanagement'

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.ROLE_MANAGEMENT, result)

    def test_get_designated_by_role(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String(Interop.GetDesignatedByRole.method_name).to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )

        path = self.get_contract_path('GetDesignatedByRole.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_get_designated_by_role_too_many_parameters(self):
        path = self.get_contract_path('GetDesignatedByRoleTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_get_designated_by_role_too_few_parameters(self):
        path = self.get_contract_path('GetDesignatedByRoleTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError, CompilerWarning
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest


class TestAny(BoaTest):
    default_folder: str = 'test_sc/any_test'

    def test_any_variable_assignments(self):
        two = String('2').to_bytes()
        path = self.get_contract_path('AnyVariableAssignments.py')

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH1      # a = 1
            + Opcode.STLOC0
            + Opcode.PUSHDATA1  # a = '2'
            + Integer(len(two)).to_byte_array() + two
            + Opcode.STLOC0
            + Opcode.PUSHT      # a = True
            + Opcode.STLOC0
            + Opcode.PUSH3      # a = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_variable_assignment_with_any(self):
        path = self.get_contract_path('VariableAssignmentWithAny.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_any_list(self):
        ok = String('ok').to_bytes()
        path = self.get_contract_path('AnyList.py')

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_any_tuple(self):
        ok = String('ok').to_bytes()
        path = self.get_contract_path('AnyTuple.py')

        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # a = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.RET        # return
        )
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_any_operation(self):
        path = self.get_contract_path('OperationWithAny.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_function_any_param(self):
        ok = String('ok').to_bytes()
        some_string = String('some_string').to_bytes()

        expected_output = (
            Opcode.INITSLOT   # Main
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHF          # bool_tuple = True, False
            + Opcode.PUSHT
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0         # SequenceFunction(bool_tuple)
            + Opcode.CALL
            + Integer(45).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHDATA1      # SequenceFunction([True, 1, 'ok'])
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.CALL
            + Integer(35).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHDATA1      # SequenceFunction('some_string')
            + Integer(len(some_string)).to_byte_array()
            + some_string
            + Opcode.CALL
            + Integer(20).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSHDATA1      # SequenceFunction((True, 1, 'ok'))
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.CALL
            + Integer(10).to_byte_array(min_length=1, signed=True)
            + Opcode.PUSH3          # SequenceFunction([1, 2, 3])
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.CALL
            + Integer(3).to_byte_array(min_length=1, signed=True)
            + Opcode.RET        # return
            + Opcode.INITSLOT   # SequenceFunction
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0         # a = sequence
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('FunctionAnyParam.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_any_sequence_assignments(self):
        ok = String('ok').to_bytes()
        some_string = String('some_string').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x05'
            + b'\x00'
            + Opcode.PUSHDATA1  # any_list = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH3      # int_list = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.PUSHDATA1  # any_tuple = (True, 1, 'ok')
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.PUSHF      # bool_tuple = True, False
            + Opcode.PUSHT
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC3
            + Opcode.LDLOC0     # a = any_list
            + Opcode.STLOC4
            + Opcode.LDLOC2     # a = any_tuple
            + Opcode.STLOC4
            + Opcode.PUSHDATA1  # a = 'some_string'
            + Integer(len(some_string)).to_byte_array() + some_string
            + Opcode.STLOC4
            + Opcode.LDLOC1     # a = int_list
            + Opcode.STLOC4
            + Opcode.LDLOC3     # a = bool_tuple
            + Opcode.STLOC4
            + Opcode.RET
        )

        path = self.get_contract_path('AnySequenceAssignments.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_int_sequence_any_assignments(self):
        ok = String('ok').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHDATA1  # any_list = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # int_sequence = any_list
            + Opcode.STLOC1
            + Opcode.RET
        )

        path = self.get_contract_path('IntSequenceAnyAssignment.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_str_sequence_any_assignments(self):
        ok = String('ok').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSHDATA1  # any_list = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.LDLOC0     # str_sequence = any_list
            + Opcode.STLOC1
            + Opcode.RET
        )

        path = self.get_contract_path('StrSequenceAnyAssignment.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_str_sequence_str_assignment(self):
        some_string = String('some_string').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSHDATA1  # str_sequence = 'some_string'
            + Integer(len(some_string)).to_byte_array() + some_string
            + Opcode.STLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('StrSequenceStrAssignment.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_sequence_of_any_sequence(self):
        ok = String('ok').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x05'
            + b'\x00'
            + Opcode.PUSHDATA1  # any_list = [True, 1, 'ok']
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH3      # int_list = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.PUSHDATA1  # any_tuple = (None, 1, 'ok')
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHNULL
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.PUSHF      # bool_tuple = True, False
            + Opcode.PUSHT
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC3
            + Opcode.LDLOC3     # a = [any_list, int_list, any_tuple, bool_tuple]
            + Opcode.LDLOC2
            + Opcode.LDLOC1
            + Opcode.LDLOC0
            + Opcode.PUSH4
            + Opcode.PACK
            + Opcode.STLOC4
            + Opcode.RET
        )

        path = self.get_contract_path('SequenceOfAnySequence.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_sequence_of_int_sequence_success(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x03'
            + b'\x00'
            + Opcode.PUSH3      # int_list = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSH8      # int_tuple = 10, 9, 8
            + Opcode.PUSH9
            + Opcode.PUSH10
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.LDLOC1     # a = [int_list, int_tuple]
            + Opcode.LDLOC0
            + Opcode.PUSH2
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.RET
        )

        path = self.get_contract_path('SequenceOfIntSequence.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_sequence_of_int_sequence_fail(self):
        ok = String('ok').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x04'
            + b'\x00'
            + Opcode.PUSH3      # int_list = [1, 2, 3]
            + Opcode.PUSH2
            + Opcode.PUSH1
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC0
            + Opcode.PUSHDATA1  # any_tuple = (True, 1, 'ok')
            + Integer(len(ok)).to_byte_array() + ok
            + Opcode.PUSH1
            + Opcode.PUSHT
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC1
            + Opcode.PUSH8      # int_tuple = 10, 9, 8
            + Opcode.PUSH9
            + Opcode.PUSH10
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC2
            + Opcode.LDLOC2     # a = [int_list, any_tuple, int_tuple]
            + Opcode.LDLOC1
            + Opcode.LDLOC0
            + Opcode.PUSH3
            + Opcode.PACK
            + Opcode.STLOC3
            + Opcode.RET
        )

        path = self.get_contract_path('SequenceOfAnyIntSequence.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)



from boa3.boa3 import Boa3
from boa3.neo import to_script_hash
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestTemplate(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = bytes(20)
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')
    OTHER_ACCOUNT_2 = bytes(range(20))

    KYC_WHITELIST_PREFIX = b'KYCWhitelistApproved'

    def test_ico_compile(self):
        path = self.get_contract_path('ico.py')
        Boa3.compile(path)

    def test_ico_verify(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'verify',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

    def test_ico_total_supply(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()
        total_supply = 10_000_000 * 10 ** 8

        result = self.run_smart_contract(engine, path, 'totalSupply',
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(total_supply, result)

    def test_ico_symbol(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'symbol')
        self.assertEqual('ICO', result)

    def test_ico_decimals(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'decimals')
        self.assertEqual(8, result)

    def test_ico_total_balance_of(self):
        total_supply = 10_000_000 * 10 ** 8

        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(total_supply, result)

        # should fail when the script length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(10))
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'balanceOf', bytes(30))

    def test_ico_kyc_register(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        # don't include if not signed by the administrator
        result = self.run_smart_contract(engine, path, 'kyc_register',
                                         [self.OWNER_SCRIPT_HASH, bytes(22)])
        self.assertEqual(0, result)

        # don't include script hashes with size different than 20
        result = self.run_smart_contract(engine, path, 'kyc_register',
                                         [bytes(40), self.OWNER_SCRIPT_HASH, bytes(12)],
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(1, result)
        storage_value = engine.storage_get(self.KYC_WHITELIST_PREFIX + self.OWNER_SCRIPT_HASH, path)
        self.assertIsNotNone(storage_value)

        # script hashes already registered are returned as well
        result = self.run_smart_contract(engine, path, 'kyc_register',
                                         [self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1],
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(2, result)
        storage_value = engine.storage_get(self.KYC_WHITELIST_PREFIX + self.OTHER_ACCOUNT_1, path)
        self.assertIsNotNone(storage_value)

    def test_ico_kyc_remove(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        # don't remove if not signed by the administrator
        result = self.run_smart_contract(engine, path, 'kyc_remove',
                                         [self.OWNER_SCRIPT_HASH, bytes(22)])
        self.assertEqual(0, result)

        # script hashes that weren't registered are returned as well
        self.assertTrue(self.KYC_WHITELIST_PREFIX + self.OTHER_ACCOUNT_1 not in engine.storage)
        result = self.run_smart_contract(engine, path, 'kyc_remove',
                                         [self.OTHER_ACCOUNT_1],
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(1, result)
        self.assertTrue(self.KYC_WHITELIST_PREFIX + self.OTHER_ACCOUNT_1 not in engine.storage)

        # don't include script hashes with size different than 20
        result = self.run_smart_contract(engine, path, 'kyc_remove',
                                         [bytes(40), self.OWNER_SCRIPT_HASH, bytes(12)],
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(1, result)

    def test_ico_approve(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        approved_amount = 100 * 10 ** 8

        # should fail when any of the scripts' length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'approve',
                                    self.OWNER_SCRIPT_HASH, bytes(10), approved_amount)
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'approve',
                                    bytes(10), self.OTHER_ACCOUNT_1, approved_amount)

        # should fail when the amount is less than 0
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'approve',
                                    self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH, -10)

        # should fail if the origin doesn't sign
        result = self.run_smart_contract(engine, path, 'approve',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, approved_amount)
        self.assertEqual(False, result)

        # should fail if origin and target are the same
        result = self.run_smart_contract(engine, path, 'approve',
                                         self.OWNER_SCRIPT_HASH, self.OWNER_SCRIPT_HASH, approved_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(False, result)

        # should fail if any of the addresses is not included in the kyc
        result = self.run_smart_contract(engine, path, 'approve',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, approved_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'kyc_register',
                                         [self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1],
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'approve',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, approved_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

    def test_ico_allowance(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        approved_amount = 100 * 10 ** 8

        result = self.run_smart_contract(engine, path, 'allowance', self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1)
        self.assertEqual(0, result)

        result = self.run_smart_contract(engine, path, 'kyc_register',
                                         [self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1],
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'approve',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, approved_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'allowance', self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1)
        self.assertEqual(approved_amount, result)

    def test_ico_transfer_from(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        transferred_amount = 100 * 10 ** 8

        # should fail when any of the scripts' length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    self.OWNER_SCRIPT_HASH, bytes(10), self.OTHER_ACCOUNT_2, transferred_amount, None)
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    bytes(10), self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2, transferred_amount, None)
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, bytes(30), transferred_amount, None)

        # should fail when the amount is less than 0
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'transferFrom',
                                    self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2, -10, None)

        # should fail if the sender doesn't sign
        result = self.run_smart_contract(engine, path, 'transferFrom',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2,
                                         transferred_amount, None)
        self.assertEqual(False, result)

        # should fail if the allowed amount is less than the given amount
        result = self.run_smart_contract(engine, path, 'transferFrom',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2,
                                         transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'kyc_register',
                                         [self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1],
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(2, result)

        result = self.run_smart_contract(engine, path, 'approve',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, transferred_amount * 2,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        # doesn't fire the transfer event when transferring to yourself or amount is zero
        balance_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        result = self.run_smart_contract(engine, path, 'transferFrom',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, self.OWNER_SCRIPT_HASH,
                                         transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        self.assertEqual(0, len(engine.get_events('transfer')))
        balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(balance_after, balance_before)

        balance_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        result = self.run_smart_contract(engine, path, 'transferFrom',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_1,
                                         transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        self.assertEqual(0, len(engine.get_events('transfer')))
        balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(balance_after, balance_before)

        result = self.run_smart_contract(engine, path, 'transferFrom',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2, 0, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(True, result)
        self.assertEqual(0, len(engine.get_events('transfer')))

        balance_originator_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        balance_sender_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        balance_receiver_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_2)
        result = self.run_smart_contract(engine, path, 'transferFrom',
                                         self.OWNER_SCRIPT_HASH, self.OTHER_ACCOUNT_1, self.OTHER_ACCOUNT_2,
                                         transferred_amount, None,
                                         signer_accounts=[self.OTHER_ACCOUNT_1])
        self.assertEqual(False, result)

        balance_originator_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(balance_originator_before, balance_originator_after)
        balance_sender_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(balance_sender_before, balance_sender_after)
        balance_receiver_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_2)
        self.assertEqual(balance_receiver_before, balance_receiver_after)

    def test_ico_mint(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()

        # should fail if amount is a negative number
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'mint', -10 * 10 ** 8)

        minted_amount = 10_000 * 10 ** 8
        # should fail if not signed by the administrator
        result = self.run_smart_contract(engine, path, 'mint', minted_amount)
        self.assertEqual(False, result)

        total_supply_before = self.run_smart_contract(engine, path, 'totalSupply')
        owner_balance_before = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)

        result = self.run_smart_contract(engine, path, 'mint', minted_amount,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertEqual(True, result)

        total_supply_after = self.run_smart_contract(engine, path, 'totalSupply')
        self.assertEqual(total_supply_before + minted_amount, total_supply_after)
        owner_balance_after = self.run_smart_contract(engine, path, 'balanceOf', self.OWNER_SCRIPT_HASH)
        self.assertEqual(owner_balance_before + minted_amount, owner_balance_after)

    def test_ico_refund(self):
        path = self.get_contract_path('ico.py')
        engine = TestEngine()
        transferred_amount = 10_000

        # should fail script hash length is not 20
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'refund', bytes(10), transferred_amount, transferred_amount)

        # should fail no amount is a positive number
        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'refund', self.OWNER_SCRIPT_HASH, 0, 0)

        # should fail if not signed by the owner
        result = self.run_smart_contract(engine, path, 'refund',
                                         self.OWNER_SCRIPT_HASH, transferred_amount, transferred_amount)
        self.assertEqual(False, result)

        # TODO: Test if the refund is successful when update the TestEngine to make Neo/Gas transfers



from typing import Dict, Optional

from boa3.model.builtin.builtinproperty import IBuiltinProperty
from boa3.model.builtin.interop.contractgethashmethod import ContractGetHashMethod
from boa3.model.variable import Variable


class GetNep17ScriptHashMethod(ContractGetHashMethod):
    def __init__(self, script_hash: bytes, identifier: str = None):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        identifier = '-get_nep17_contract' if not isinstance(identifier, str) else identifier
        args: Dict[str, Variable] = {}
        super().__init__(script_hash, identifier, args, return_type=UInt160Type.build())

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None


class Nep17ContractProperty(IBuiltinProperty):
    def __init__(self, script_hash: bytes):
        identifier = 'Nep17'
        getter = GetNep17ScriptHashMethod(script_hash)
        super().__init__(identifier, getter)


Nep17Contract = Nep17ContractProperty(b'')



import argparse
import logging
import os
import sys

from boa3.boa3 import Boa3
from boa3.exception.NotLoadedException import NotLoadedException


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("input", help=".py smart contract to compile")
    parser.add_argument("-db", "--debug", action='store_true', help="generates a .nefdbgnfo file")
    parser.add_argument("--project-path", help="Project root path. Path of the contract by default.", type=str)
    args = parser.parse_args()

    if not args.input.endswith(".py") or not os.path.isfile(args.input):
        logging.error("Input file is not .py")
        sys.exit(1)

    fullpath = os.path.realpath(args.input)
    path, filename = os.path.split(fullpath)

    if not args.project_path:
        args.project_path = os.path.dirname(path)

    try:
        Boa3.compile_and_save(args.input, debug=args.debug, root_folder=args.project_path)
        logging.info(f"Wrote {filename.replace('.py', '.nef')} to {path}")
    except NotLoadedException as e:
        error_message = e.message
        log_error = 'Could not compile'
        if len(error_message) > 0:
            log_error += f': {error_message}'

        logging.error(log_error)
    except Exception as e:
        logging.exception(e)


if __name__ == "__main__":
    main()



from typing import Any, Sequence

from boa3.neo.vm.type.AbiType import AbiType


def value_to_parameter(value: Any) -> Any:
    if isinstance(value, (int, str)):
        return value

    if isinstance(value, bytes):
        return bytes_to_hex(value)

    if isinstance(value, Sequence):
        return [value_to_parameter(x) for x in value]

    if isinstance(value, dict):
        return {
            'type': AbiType.Map.value,
            'value': [{'key': value_to_parameter(key),
                       'value': value_to_parameter(value)
                       } for key, value in value.items()
                      ]
        }

    if value is not None:
        raise TypeError(f"Unsupported type: '{type(value)}'")

    return None


def bytes_to_hex(data: bytes) -> str:
    return '0x' + data.hex()



from __future__ import annotations

from typing import Union

from boa3 import constants
from boa3.builtin.compile_time import NeoMetadata
from boa3.neo3.core.types import UInt160


class CompiledMetadata:
    _instance = None

    @classmethod
    def instance(cls) -> CompiledMetadata:
        if cls._instance is None:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self._metadata = NeoMetadata()

    @classmethod
    def reset(cls):
        cls.instance()._metadata = NeoMetadata()

    @classmethod
    def set_current_metadata(cls, metadata: NeoMetadata):
        cls.instance()._metadata = metadata

    def add_contract_permission(self, contract: Union[bytes, str], method: str = None):
        if isinstance(contract, bytes):
            contract = UInt160(contract)
        else:
            contract = UInt160.from_string(contract)
        contract_hex_script = str(contract)
        method_string = method if isinstance(method, str) and len(method) > 0 else constants.IMPORT_WILDCARD

        # look for permissions for the given contract
        existing_permission = next((permission for permission in self._metadata.permissions
                                    if ('contract' in permission
                                        and permission['contract'] == contract_hex_script)
                                    ), None)

        if existing_permission is not None:
            # if there's already a permission, include the method name
            permitted_methods = existing_permission['methods'] if 'methods' in existing_permission else []

            # if it's using the wildcard, all methods in this contract can already be called
            if permitted_methods != constants.IMPORT_WILDCARD:
                if method_string == constants.IMPORT_WILDCARD:
                    existing_permission['methods'] = constants.IMPORT_WILDCARD
                else:
                    if method_string not in permitted_methods:
                        permitted_methods.append(method_string)

                    if 'methods' not in existing_permission:
                        existing_permission['methods'] = permitted_methods

            return

        # look for generic permissions for the given method
        existing_permission = next((permission for permission in self._metadata.permissions
                                    if ('contract' in permission
                                        and permission['contract'] == constants.IMPORT_WILDCARD
                                        and 'method' in permission
                                        and (permission['methods'] == constants.IMPORT_WILDCARD
                                             or method_string in permission['methods']))
                                    ), None)

        if existing_permission is None:
            # if no permission is found, include it
            self._metadata.add_permission(contract=contract_hex_script,
                                          methods=([method_string] if method_string != constants.IMPORT_WILDCARD
                                                   else method_string))



from typing import Dict

from boa3.model.builtin.interop.contract.callflagstype import CallFlagsType
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class GetCallFlagsMethod(InteropMethod):

    def __init__(self, call_flags_type: CallFlagsType):
        identifier = 'get_call_flags'
        syscall = 'System.Contract.GetCallFlags'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=call_flags_type)



from typing import Dict

from boa3.model.builtin.interop.nativecontract import Nep17Method
from boa3.model.variable import Variable


class TotalSupplyMethod(Nep17Method):

    def __init__(self, contract_script_hash: bytes):
        from boa3.model.type.type import Type

        identifier = 'totalSupply'
        native_identifier = 'totalSupply'
        args: Dict[str, Variable] = {}
        super().__init__(identifier, native_identifier, args, return_type=Type.int, script_hash=contract_script_hash)




import boa3_test.test_sc.import_test.variable_import.StaticVariables as StaticVariables
from boa3.builtin.compile_time import public


@public
def get_foo() -> bytes:
    return StaticVariables.FOO


@public
def get_bar() -> str:
    return StaticVariables.BAR



from boa3 import constants
from boa3.exception import CompilerError
from boa3.neo import from_hex_str
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestNeoClass(BoaTest):
    default_folder: str = 'test_sc/native_test/neo'

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.NEO_SCRIPT, result)

    def test_symbol(self):
        path = self.get_contract_path('Symbol.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual('NEO', result)

    def test_symbol_too_many_parameters(self):
        path = self.get_contract_path('SymbolTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_decimals(self):
        path = self.get_contract_path('Decimals.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(0, result)

    def test_decimals_too_many_parameters(self):
        path = self.get_contract_path('DecimalsTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_total_supply(self):
        path = self.get_contract_path('TotalSupply.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(100_000_000, result)

    def test_total_supply_too_many_parameters(self):
        path = self.get_contract_path('TotalSupplyTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_balance_of(self):
        path = self.get_contract_path('BalanceOf.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', bytes(range(20)))
        self.assertEqual(0, result)

        engine.add_neo(bytes(range(20)), 10)
        result = self.run_smart_contract(engine, path, 'main', bytes(range(20)))
        self.assertEqual(10, result)

    def test_balance_of_too_many_parameters(self):
        path = self.get_contract_path('BalanceOfTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_transfer(self):
        path = self.get_contract_path('Transfer.py')
        engine = TestEngine()

        account_1 = bytes(range(20))
        account_2 = bytes(range(20)[::-1])
        amount = 10000

        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount, ['value', 123, False])
        self.assertEqual(False, result)

        engine.add_neo(account_1, amount)
        # can't transfer if there is no signature, even with enough GAS
        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount, ['value', 123, False])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount, ['value', 123, False],
                                         signer_accounts=[account_1])
        self.assertEqual(True, result)

    def test_transfer_data_default(self):
        path = self.get_contract_path('TransferDataDefault.py')
        engine = TestEngine()

        account_1 = bytes(range(20))
        account_2 = bytes(range(20)[::-1])
        amount = 100

        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount)
        self.assertEqual(False, result)

        engine.add_neo(account_1, amount)
        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount,
                                         signer_accounts=[account_1])
        self.assertEqual(True, result)

    def test_transfer_too_many_parameters(self):
        path = self.get_contract_path('TransferTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_transfer_too_few__parameters(self):
        path = self.get_contract_path('TransferTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_get_gas_per_block(self):
        path = self.get_contract_path('GetGasPerBlock.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(5 * 10 ** 8, result)

    def test_unclaimed_gas(self):
        path = self.get_contract_path('UnclaimedGas.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', bytes(20), 0)
        self.assertIsInstance(result, int)

    def test_register_candidate(self):
        path = self.get_contract_path('RegisterCandidate.py')
        engine = TestEngine()

        candidate_pubkey = bytes.fromhex('0296852e74830f48185caec9980d21dee5e8bee3da97d712123c19ee01c2d3f3ae')
        candidate_scripthash = from_hex_str('a8de26eb4931c674d31885acf722bd82e6bcd06d')
        result = self.run_smart_contract(engine, path, 'main', candidate_pubkey)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', candidate_pubkey,
                                         signer_accounts=[candidate_scripthash])
        self.assertEqual(True, result)

    def test_unregister_candidate(self):
        path = self.get_contract_path('UnregisterCandidate.py')
        engine = TestEngine()

        candidate_pubkey = bytes.fromhex('0296852e74830f48185caec9980d21dee5e8bee3da97d712123c19ee01c2d3f3ae')
        candidate_scripthash = from_hex_str('a8de26eb4931c674d31885acf722bd82e6bcd06d')
        result = self.run_smart_contract(engine, path, 'main', candidate_pubkey)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', candidate_pubkey, signer_accounts=[candidate_scripthash])
        self.assertEqual(True, result)

    def test_vote(self):
        path = self.get_contract_path('Vote.py')
        engine = TestEngine()

        account = bytes(range(20))
        n_votes = 100
        candidate_pubkey = bytes.fromhex('0296852e74830f48185caec9980d21dee5e8bee3da97d712123c19ee01c2d3f3ae')
        # will fail check_witness
        result = self.run_smart_contract(engine, path, 'main', account, candidate_pubkey)
        self.assertEqual(False, result)

        # NeoAccountState is None and will return false
        result = self.run_smart_contract(engine, path, 'main', account, candidate_pubkey, signer_accounts=[account])
        self.assertEqual(False, result)

        # adding NEO to the account will make NeoAccountState not None
        engine.add_neo(account, n_votes)

        # candidate is not registered yet
        result = self.run_smart_contract(engine, path, 'main', account, candidate_pubkey, signer_accounts=[account])
        self.assertEqual(False, result)

        path_register = self.get_contract_path('RegisterCandidate.py')
        candidate_pubkey = bytes.fromhex('0296852e74830f48185caec9980d21dee5e8bee3da97d712123c19ee01c2d3f3ae')
        candidate_scripthash = from_hex_str('a8de26eb4931c674d31885acf722bd82e6bcd06d')
        self.run_smart_contract(engine, path_register, 'main', candidate_pubkey,
                                signer_accounts=[candidate_scripthash])

        # candidate was registered
        result = self.run_smart_contract(engine, path, 'main', account, candidate_pubkey, signer_accounts=[account])
        self.assertEqual(True, result)

        path_get = self.get_contract_path('GetCandidates.py')
        result = self.run_smart_contract(engine, path_get, 'main')
        self.assertEqual(1, len(result))
        self.assertEqual(candidate_pubkey, result[0][0])
        self.assertEqual(n_votes, result[0][1])

        # remove votes from candidate
        result = self.run_smart_contract(engine, path, 'un_vote', account, signer_accounts=[account])
        self.assertEqual(True, result)

        # candidate has no votes now
        result = self.run_smart_contract(engine, path_get, 'main')
        self.assertEqual(1, len(result))
        self.assertEqual(candidate_pubkey, result[0][0])
        self.assertEqual(0, result[0][1])

        # voting for candidate again
        self.run_smart_contract(engine, path, 'main', account, candidate_pubkey, signer_accounts=[account])
        result = self.run_smart_contract(engine, path_get, 'main')
        self.assertEqual(n_votes, result[0][1])

        # it's also possible to remove votes by voting for None instead of using un_vote
        result = self.run_smart_contract(engine, path, 'main', account, None, signer_accounts=[account])
        self.assertEqual(True, result)

        # candidate has no votes now again
        result = self.run_smart_contract(engine, path_get, 'main')
        self.assertEqual(1, len(result))
        self.assertEqual(candidate_pubkey, result[0][0])
        self.assertEqual(0, result[0][1])

    def test_un_vote_too_many_parameters(self):
        path = self.get_contract_path('UnVoteTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_un_vote_too_few_parameters(self):
        path = self.get_contract_path('UnVoteTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_get_all_candidates(self):
        path = self.get_contract_path('GetAllCandidates.py')
        self.compile_and_save(path)
        engine = TestEngine()

        # no candidate was registered
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(0, len(result))

        path_register = self.get_contract_path('RegisterCandidate.py')
        candidate_pubkey = bytes.fromhex('0296852e74830f48185caec9980d21dee5e8bee3da97d712123c19ee01c2d3f3ae')
        candidate_scripthash = from_hex_str('a8de26eb4931c674d31885acf722bd82e6bcd06d')
        result = self.run_smart_contract(engine, path_register, 'main', candidate_pubkey,
                                         signer_accounts=[candidate_scripthash])
        self.assertTrue(result)

        # after registering one
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(1, len(result))
        self.assertEqual(candidate_pubkey, result[0][0])
        self.assertEqual(0, result[0][1])

    def test_get_candidates(self):
        path = self.get_contract_path('GetCandidates.py')
        engine = TestEngine()

        # no candidate was registered
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(0, len(result))

        path_register = self.get_contract_path('RegisterCandidate.py')
        candidate_pubkey = bytes.fromhex('0296852e74830f48185caec9980d21dee5e8bee3da97d712123c19ee01c2d3f3ae')
        candidate_scripthash = from_hex_str('a8de26eb4931c674d31885acf722bd82e6bcd06d')
        self.run_smart_contract(engine, path_register, 'main', candidate_pubkey,
                                signer_accounts=[candidate_scripthash])

        # after registering one
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(1, len(result))
        self.assertEqual(candidate_pubkey, result[0][0])
        self.assertEqual(0, result[0][1])

    def test_get_candidate_vote(self):
        path = self.get_contract_path('GetCandidateVote.py')
        engine = TestEngine()

        account = bytes(range(20))
        n_votes = 100
        candidate_pubkey = bytes.fromhex('0296852e74830f48185caec9980d21dee5e8bee3da97d712123c19ee01c2d3f3ae')
        candidate_scripthash = from_hex_str('a8de26eb4931c674d31885acf722bd82e6bcd06d')
        # will fail check_witness
        result = self.run_smart_contract(engine, path, 'main', candidate_pubkey)
        self.assertEqual(-1, result)

        engine.add_neo(account, n_votes)
        path_register = self.get_contract_path('RegisterCandidate.py')
        result = self.run_smart_contract(engine, path_register, 'main', candidate_pubkey,
                                         signer_accounts=[candidate_scripthash])
        self.assertTrue(result)

        # candidate was registered
        path_vote = self.get_contract_path('Vote.py')
        result = self.run_smart_contract(engine, path_vote, 'main', account, candidate_pubkey,
                                         signer_accounts=[account])
        self.assertTrue(result)

        result = self.run_smart_contract(engine, path, 'main', candidate_pubkey)
        self.assertEqual(n_votes, result)

    def test_get_committee(self):
        path = self.get_contract_path('GetCommittee.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        default_council = [
            bytes.fromhex('03b209fd4f53a7170ea4444e0cb0a6bb6a53c2bd016926989cf85f9b0fba17a70c'),
            bytes.fromhex('02486fd15702c4490a26703112a5cc1d0923fd697a33406bd5a1c00e0013b09a70'),
            bytes.fromhex('02ca0e27697b9c248f6f16e085fd0061e26f44da85b58ee835c110caa5ec3ba554'),
            bytes.fromhex('024c7b7fb6c310fccf1ba33b082519d82964ea93868d676662d4a59ad548df0e7d'),
            bytes.fromhex('03b8d9d5771d8f513aa0869b9cc8d50986403b78c6da36890638c3d46a5adce04a'),
            bytes.fromhex('02df48f60e8f3e01c48ff40b9b7f1310d7a8b2a193188befe1c2e3df740e895093'),
            bytes.fromhex('02aaec38470f6aad0042c6e877cfd8087d2676b0f516fddd362801b9bd3936399e'),
            bytes.fromhex('023a36c72844610b4d34d1968662424011bf783ca9d984efa19a20babf5582f3fe'),
            bytes.fromhex('03708b860c1de5d87f5b151a12c2a99feebd2e8b315ee8e7cf8aa19692a9e18379'),
            bytes.fromhex('03c6aa6e12638b36e88adc1ccdceac4db9929575c3e03576c617c49cce7114a050'),
            bytes.fromhex('03204223f8c86b8cd5c89ef12e4f0dbb314172e9241e30c9ef2293790793537cf0'),
            bytes.fromhex('02a62c915cf19c7f19a50ec217e79fac2439bbaad658493de0c7d8ffa92ab0aa62'),
            bytes.fromhex('03409f31f0d66bdc2f70a9730b66fe186658f84a8018204db01c106edc36553cd0'),
            bytes.fromhex('0288342b141c30dc8ffcde0204929bb46aed5756b41ef4a56778d15ada8f0c6654'),
            bytes.fromhex('020f2887f41474cfeb11fd262e982051c1541418137c02a0f4961af911045de639'),
            bytes.fromhex('0222038884bbd1d8ff109ed3bdef3542e768eef76c1247aea8bc8171f532928c30'),
            bytes.fromhex('03d281b42002647f0113f36c7b8efb30db66078dfaaa9ab3ff76d043a98d512fde'),
            bytes.fromhex('02504acbc1f4b3bdad1d86d6e1a08603771db135a73e61c9d565ae06a1938cd2ad'),
            bytes.fromhex('0226933336f1b75baa42d42b71d9091508b638046d19abd67f4e119bf64a7cfb4d'),
            bytes.fromhex('03cdcea66032b82f5c30450e381e5295cae85c5e6943af716cc6b646352a6067dc'),
            bytes.fromhex('02cd5a5547119e24feaa7c2a0f37b8c9366216bab7054de0065c9be42084003c8a'),
        ]
        is_committee_member = True
        for pubkey in default_council:
            if pubkey not in result:
                is_committee_member = False
        self.assertEqual(True, is_committee_member)

    def test_get_next_block_validators(self):
        path = self.get_contract_path('GetNextBlockValidators.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        consensus_nodes = [
            bytes.fromhex('03b209fd4f53a7170ea4444e0cb0a6bb6a53c2bd016926989cf85f9b0fba17a70c'),
            bytes.fromhex('02486fd15702c4490a26703112a5cc1d0923fd697a33406bd5a1c00e0013b09a70'),
            bytes.fromhex('02ca0e27697b9c248f6f16e085fd0061e26f44da85b58ee835c110caa5ec3ba554'),
            bytes.fromhex('024c7b7fb6c310fccf1ba33b082519d82964ea93868d676662d4a59ad548df0e7d'),
            bytes.fromhex('03b8d9d5771d8f513aa0869b9cc8d50986403b78c6da36890638c3d46a5adce04a'),
            bytes.fromhex('02df48f60e8f3e01c48ff40b9b7f1310d7a8b2a193188befe1c2e3df740e895093'),
            bytes.fromhex('02aaec38470f6aad0042c6e877cfd8087d2676b0f516fddd362801b9bd3936399e'),
        ]
        is_consensus_node = True
        for pubkey in consensus_nodes:
            if pubkey not in result:
                is_consensus_node = False
        self.assertEqual(True, is_consensus_node)

    def test_get_account_state(self):
        path = self.get_contract_path('GetAccountState.py')
        engine = TestEngine()

        account = bytes(range(20))
        result = self.run_smart_contract(engine, path, 'main', account)
        self.assertIsNone(result)

        # adding votes
        votes = 10000
        engine.add_neo(account, votes)
        result = self.run_smart_contract(engine, path, 'main', account)
        self.assertEqual(3, len(result))
        # number of votes in the account
        self.assertEqual(votes, result[0])
        # balance was changed at height 0
        self.assertEqual(0, result[1])
        # who the account is voting for
        self.assertIsNone(result[2])

        engine.increase_block(10)
        other_account = bytes(20)

        path_transfer = self.get_contract_path('Transfer.py')

        self.run_smart_contract(engine, path_transfer, 'main', account, other_account, 1, None,
                                signer_accounts=[account])
        votes = votes - 1

        result = self.run_smart_contract(engine, path, 'main', account)
        self.assertEqual(3, len(result))
        self.assertEqual(votes, result[0])
        self.assertEqual(10, result[1])
        self.assertIsNone(result[2])



from boa3.exception import CompilerError
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestBuiltinMethod(BoaTest):
    default_folder: str = 'test_sc/boa_built_in_methods_test'

    def test_abort(self):
        path = self.get_contract_path('Abort.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', False)
        self.assertEqual(123, result)

        with self.assertRaisesRegex(TestExecutionException, self.ABORTED_CONTRACT_MSG):
            self.run_smart_contract(engine, path, 'main', True)

    def test_deploy_def(self):
        path = self.get_contract_path('DeployDef.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'get_var')
        self.assertEqual(10, result)

    def test_deploy_def_incorrect_signature(self):
        path = self.get_contract_path('DeployDefWrongSignature.py')
        self.assertCompilerLogs(CompilerError.InternalIncorrectSignature, path)

    def test_will_not_compile(self):
        path = self.get_contract_path('WillNotCompile.py')
        self.assertCompilerLogs(CompilerError.UnresolvedReference, path)

    # region math builtins

    def test_sqrt_method(self):
        path = self.get_contract_path('Sqrt.py')
        engine = TestEngine()

        from math import sqrt

        expected_result = int(sqrt(0))
        result = self.run_smart_contract(engine, path, 'main', 0)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(1))
        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(3))
        result = self.run_smart_contract(engine, path, 'main', 3)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(4))
        result = self.run_smart_contract(engine, path, 'main', 4)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(8))
        result = self.run_smart_contract(engine, path, 'main', 8)
        self.assertEqual(expected_result, result)
        expected_result = int(sqrt(10))
        result = self.run_smart_contract(engine, path, 'main', 10)
        self.assertEqual(expected_result, result)

        with self.assertRaisesRegex(TestExecutionException, self.VALUE_CANNOT_BE_NEGATIVE_MSG):
            self.run_smart_contract(engine, path, 'main', -1)

        val = 25
        expected_result = int(sqrt(val))
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(expected_result, result)

    def test_sqrt_method_from_math(self):
        path = self.get_contract_path('SqrtFromMath.py')
        engine = TestEngine()

        from math import sqrt

        val = 25
        expected_result = int(sqrt(val))
        result = self.run_smart_contract(engine, path, 'main', val)
        self.assertEqual(expected_result, result)

    def test_decimal_floor_method(self):
        path = self.get_contract_path('DecimalFloor.py')
        engine = TestEngine()

        from math import floor

        value = 4.2
        decimals = 8

        multiplier = 10 ** decimals
        value_floor = int(floor(value)) * multiplier
        integer_value = int(value * multiplier)
        result = self.run_smart_contract(engine, path, 'main', integer_value, decimals)
        self.assertEqual(value_floor, result)

        decimals = 12

        multiplier = 10 ** decimals
        value_floor = int(floor(value)) * multiplier
        integer_value = int(value * multiplier)
        result = self.run_smart_contract(engine, path, 'main', integer_value, decimals)
        self.assertEqual(value_floor, result)

        value = -3.983541

        multiplier = 10 ** decimals
        value_floor = int(floor(value) * multiplier)
        integer_value = int(value * multiplier)
        result = self.run_smart_contract(engine, path, 'main', integer_value, decimals)
        self.assertEqual(value_floor, result)

        # negative decimals will raise an exception
        from boa3.model.builtin.builtin import Builtin
        with self.assertRaisesRegex(TestExecutionException, f'{Builtin.BuiltinMathFloor.exception_message}$'):
            self.run_smart_contract(engine, path, 'main', integer_value, -1)

    def test_decimal_ceil_method(self):
        path = self.get_contract_path('DecimalCeiling.py')
        engine = TestEngine()

        from math import ceil

        value = 4.2
        decimals = 8

        multiplier = 10 ** decimals
        value_ceiling = int(ceil(value)) * multiplier
        integer_value = int(value * multiplier)
        result = self.run_smart_contract(engine, path, 'main', integer_value, decimals)
        self.assertEqual(value_ceiling, result)

        decimals = 12

        multiplier = 10 ** decimals
        value_ceiling = int(ceil(value)) * multiplier
        integer_value = int(value * multiplier)
        result = self.run_smart_contract(engine, path, 'main', integer_value, decimals)
        self.assertEqual(value_ceiling, result)

        value = -3.983541

        multiplier = 10 ** decimals
        value_ceiling = int(ceil(value) * multiplier)
        integer_value = int(value * multiplier)
        result = self.run_smart_contract(engine, path, 'main', integer_value, decimals)
        self.assertEqual(value_ceiling, result)

        # negative decimals will raise an exception
        from boa3.model.builtin.builtin import Builtin
        with self.assertRaisesRegex(TestExecutionException, f'{Builtin.BuiltinMathCeil.exception_message}$'):
            self.run_smart_contract(engine, path, 'main', integer_value, -1)

    # endregion



import ast
from typing import Any, Dict, List, Optional, Tuple

from boa3.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class ListMethod(IBuiltinMethod):

    def __init__(self, args: Dict[str, Variable] = None, return_type=None, defaults: List[ast.AST] = None):
        identifier = 'list'

        self._prepare_for_packing = None

        super().__init__(identifier, args, defaults=defaults, return_type=return_type)

    @property
    def _arg_value(self) -> Variable:
        return self.args['value']

    @property
    def identifier(self) -> str:
        from boa3.model.type.type import Type

        if self._arg_value.type is Type.str or self._arg_value.type is Type.bytes:
            return '-{0}_{1}'.format(self._identifier, Type.str.identifier + Type.bytes.identifier)

        if Type.sequence.is_type_of(self._arg_value.type):
            return '-{0}_{1}'.format(self._identifier, Type.sequence.identifier)

        if Type.mapping.is_type_of(self._arg_value.type):
            return '-{0}_{1}'.format(self._identifier, Type.mapping.identifier)

        return self._identifier

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:

        pack_values = [(Opcode.PACK, b'')]

        return self.prepare_for_packing + pack_values

    @property
    def prepare_for_packing(self) -> List[Tuple[Opcode, bytes]]:

        if self._prepare_for_packing is None:
            self._prepare_for_packing = []

        return self._prepare_for_packing

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> Optional[str]:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        from boa3.model.type.type import Type

        from boa3.model.type.annotation.uniontype import UnionType
        is_union = len(value) > 0 and isinstance(value[0], UnionType)
        union_types = value[0].union_types if is_union else None

        if len(value) > 0:
            if Type.str.is_type_of(value[0]) or Type.bytes.is_type_of(value[0]) or \
                    (is_union and all((union_type is Type.str or union_type is Type.bytes) for union_type in union_types)):
                from boa3.model.builtin.method import ListBytesStringMethod
                return ListBytesStringMethod(value[0])

            if Type.mapping.is_type_of(value[0]):
                from boa3.model.builtin.method import ListMappingMethod
                return ListMappingMethod(value[0])

            if is_union and \
                    any((union_type is Type.str or
                         union_type is Type.bytes or
                         Type.sequence.is_type_of(union_type) or
                         Type.mapping.is_type_of(union_type)) for union_type in union_types):
                from boa3.model.builtin.method import ListGenericMethod
                return ListGenericMethod(value[0])

            if value[0] is Type.any:
                from boa3.model.builtin.method import ListGenericMethod
                return ListGenericMethod()

            if Type.sequence.is_type_of(value[0]):
                from boa3.model.builtin.method import ListSequenceMethod
                return ListSequenceMethod(value[0])



from boa3 import constants
from boa3.exception import CompilerError
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestGasClass(BoaTest):
    default_folder: str = 'test_sc/native_test/gas'

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.GAS_SCRIPT, result)

    def test_symbol(self):
        path = self.get_contract_path('Symbol.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual('GAS', result)

    def test_symbol_too_many_parameters(self):
        path = self.get_contract_path('SymbolTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_decimals(self):
        path = self.get_contract_path('Decimals.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(8, result)

    def test_decimals_too_many_parameters(self):
        path = self.get_contract_path('DecimalsTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_total_supply(self):
        path = self.get_contract_path('TotalSupply.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsInstance(result, int)

    def test_total_supply_too_many_parameters(self):
        path = self.get_contract_path('TotalSupplyTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_balance_of(self):
        path = self.get_contract_path('BalanceOf.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', bytes(range(20)))
        self.assertEqual(0, result)

        engine.add_gas(bytes(range(20)), 10)
        result = self.run_smart_contract(engine, path, 'main', bytes(range(20)))
        self.assertEqual(10, result)

    def test_balance_of_too_many_parameters(self):
        path = self.get_contract_path('BalanceOfTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_transfer(self):
        path = self.get_contract_path('Transfer.py')
        engine = TestEngine()

        account_1 = bytes(range(20))
        account_2 = bytes(range(20)[::-1])
        amount = 10000

        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount, ['value', 123, False])
        self.assertEqual(False, result)

        engine.add_gas(account_1, amount)
        # can't transfer if there is no signature, even with enough GAS
        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount, ['value', 123, False])
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount, ['value', 123, False],
                                         signer_accounts=[account_1])
        self.assertEqual(True, result)

    def test_transfer_data_default(self):
        path = self.get_contract_path('TransferDataDefault.py')
        engine = TestEngine()

        account_1 = bytes(range(20))
        account_2 = bytes(range(20)[::-1])
        amount = 100

        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount)
        self.assertEqual(False, result)

        engine.add_gas(account_1, amount)
        result = self.run_smart_contract(engine, path, 'main', account_1, account_2, amount,
                                         signer_accounts=[account_1])
        self.assertEqual(True, result)

    def test_transfer_too_many_parameters(self):
        path = self.get_contract_path('TransferTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_transfer_too_few__parameters(self):
        path = self.get_contract_path('TransferTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_import_with_alias(self):
        path = self.get_contract_path('ImportWithAlias.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', bytes(20))
        self.assertEqual(0, result)



from __future__ import annotations

import abc
from typing import Any, Dict

from boa3_test.tests.test_classes import transactionattribute


class TransactionAttribute(abc.ABC):

    def __init__(self, _type: transactionattribute.TransactionAttributeType):
        self._type: transactionattribute.TransactionAttributeType = _type

    def to_json(self) -> Dict[str, Any]:
        return {
            'type': self._type.name
        }

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> TransactionAttribute:
        tx_type = json['type']
        pass

    def __eq__(self, other) -> bool:
        return isinstance(other, type(self)) and self._type == other._type



from typing import Dict

from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.variable import Variable


class CheckMultisigMethod(InteropMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'check_multisig'
        syscall = 'System.Crypto.CheckMultisig'
        args: Dict[str, Variable] = {
            'pubkeys': Variable(Type.list.build_collection([ECPointType.build()])),
            'signatures': Variable(Type.list.build_collection([Type.bytes]))
        }
        super().__init__(identifier, syscall, args, return_type=Type.bool)



import json

from boa3.boa3 import Boa3
from boa3.neo import to_script_hash
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestUpdateContractTemplate(BoaTest):
    default_folder: str = 'examples'

    OWNER_SCRIPT_HASH = bytes(20)
    OTHER_ACCOUNT_1 = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')

    def test_update_contract_compile(self):
        path = self.get_contract_path('update_contract.py')
        path_new = self.get_contract_path('examples/auxiliary_contracts', 'update_contract.py')
        Boa3.compile(path)
        Boa3.compile(path_new)

    def test_update_contract(self):
        path = self.get_contract_path('update_contract.py')
        path_new = self.get_contract_path('examples/auxiliary_contracts', 'update_contract.py')
        self.get_output(path_new)

        new_nef, new_manifest = self.get_bytes_output(path_new)
        arg_manifest = String(json.dumps(new_manifest, separators=(',', ':'))).to_bytes()
        engine = TestEngine()

        # Saving user's balance before calling method to compare it later
        tokens_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)

        event_transfer = engine.get_events('Transfer')
        # Transfer emitted when deploying the smart contract
        self.assertEqual(1, len(event_transfer))

        # The bugged method is being called and the user is able to receive tokens for free
        result = self.run_smart_contract(engine, path, 'method', self.OTHER_ACCOUNT_1)
        self.assertIsVoid(result)

        event_transfer = engine.get_events('Transfer')
        self.assertEqual(2, len(event_transfer))
        # The amount of tokens will be higher after calling the method
        tokens_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertGreater(tokens_after, tokens_before)

        # The smart contract will be updated to fix the bug in the method
        result = self.run_smart_contract(engine, path, 'update_sc', new_nef, arg_manifest, None,
                                         signer_accounts=[self.OWNER_SCRIPT_HASH])
        self.assertIsVoid(result)

        # An `Update` event will be emitted
        event_update = engine.get_events('Update')
        self.assertEqual(1, len(event_update))

        # Saving user's balance before calling method to compare it later
        tokens_before = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)

        # Now, when method is called, it won't mint new tokens to any user that called it
        result = self.run_smart_contract(engine, path, 'method', self.OTHER_ACCOUNT_1)
        self.assertIsVoid(result)

        # The amount of tokens now is the same before and after calling the method
        tokens_after = self.run_smart_contract(engine, path, 'balanceOf', self.OTHER_ACCOUNT_1)
        self.assertEqual(tokens_after, tokens_before)



from boa3.exception import CompilerError, CompilerWarning
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestTyping(BoaTest):
    default_folder: str = 'test_sc/typing_test'

    def test_cast_to_int(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = cast(int, value)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('CastToInt.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_cast_to_str(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = cast(str, value)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('CastToStr.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_cast_to_list(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = cast(list, value)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('CastToList.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_cast_to_typed_list(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = cast(List[int], value)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return x[0]
            + Opcode.PUSH0
            + Opcode.DUP
            + Opcode.SIGN
            + Opcode.PUSHM1
            + Opcode.JMPNE
            + Integer(5).to_byte_array(min_length=1, signed=True)
            + Opcode.OVER
            + Opcode.SIZE
            + Opcode.ADD
            + Opcode.PICKITEM
            + Opcode.RET
        )

        path = self.get_contract_path('CastToTypedList.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_cast_to_dict(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = cast(dict, value)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('CastToDict.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_cast_to_typed_dict(self):
        string = String('example').to_bytes()
        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = cast(Dict[str, int], value)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return x['example']
            + Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.PICKITEM
            + Opcode.RET
        )

        path = self.get_contract_path('CastToTypedDict.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_cast_mismatched_type(self):
        path = self.get_contract_path('CastMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_cast_to_uint160(self):
        path = self.get_contract_path('CastToUInt160.py')
        self.assertCompilerLogs(CompilerWarning.TypeCasting, path)

        engine = TestEngine()
        value = bytes(range(20))
        result = self.run_smart_contract(engine, path, 'Main', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

    def test_cast_to_transaction(self):
        expected_output = (
            Opcode.INITSLOT     # function signature
            + b'\x01'
            + b'\x01'
            + Opcode.LDARG0     # x = cast(Transaction, value)
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return x
            + Opcode.RET
        )

        path = self.get_contract_path('CastToTransaction.py')
        output = self.assertCompilerLogs(CompilerWarning.TypeCasting, path)
        self.assertEqual(expected_output, output)

    def test_cast_inside_if(self):
        path = self.get_contract_path('CastInsideIf.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual('body', result)

    def test_cast_persisted_in_scope(self):
        path = self.get_contract_path('CastPersistedInScope.py')
        engine = TestEngine()

        test_address = bytes(20)
        result = self.run_smart_contract(engine, path, 'main', test_address, 10, None)
        self.assertIsVoid(result)



from typing import Any, List, Tuple

from boa3.builtin.contract import NeoAccountState
from boa3.builtin.interop.iterator import Iterator
from boa3.builtin.type import ECPoint, UInt160


class NEO:
    """
    A class used to represent the NEO native contract
    """

    hash: UInt160

    @classmethod
    def symbol(cls) -> str:
        """
        Gets the symbol of NEO.

        :return: the NEO string.
        :rtype: str
        """
        pass

    @classmethod
    def decimals(cls) -> int:
        """
        Gets the amount of decimals used by NEO.

        :return: the number 8.
        :rtype: int
        """
        pass

    @classmethod
    def totalSupply(cls) -> int:
        """
        Gets the total token supply deployed in the system.

        :return: the total token supply deployed in the system.
        :rtype: int
        """
        pass

    @classmethod
    def balanceOf(cls, account: UInt160) -> int:
        """
        Get the current balance of an address.

        :param account: the account's address to retrieve the balance for
        :type account: UInt160
        :return: the account's balance
        :rtype: int
        """
        pass

    @classmethod
    def transfer(cls, from_address: UInt160, to_address: UInt160, amount: int, data: Any = None) -> bool:
        """
        Transfers an amount of GAS from one account to another

        If the method succeeds, it will fire the `Transfer` event and must return true, even if the amount is 0,
        or from and to are the same address.

        :param from_address: the address to transfer from
        :type from_address: UInt160
        :param to_address: the address to transfer to
        :type to_address: UInt160
        :param amount: the amount of NEO to transfer
        :type amount: int
        :param data: whatever data is pertinent to the onNEP17Payment method
        :type data: Any

        :return: whether the transfer was successful
        :rtype: bool
        :raise Exception: raised if `from_address` or `to_address` length is not 20 or if `amount` is less than zero.
        """
        pass

    @classmethod
    def get_gas_per_block(cls) -> int:
        """
        Gets the amount of GAS generated in each block.

        :return: the amount of GAS generated
        :rtype: int
        """
        pass

    @classmethod
    def unclaimed_gas(cls, account: UInt160, end: int) -> int:
        """
        Gets the amount of unclaimed GAS in the specified account.

        :param account: the account to check
        :type account: UInt160
        :param end: the block index used when calculating GAS
        :type end: int
        """
        pass

    @classmethod
    def register_candidate(cls, pubkey: ECPoint) -> bool:
        """
        Registers as a candidate.

        :param pubkey: The public key of the account to be registered
        :type pubkey: ECPoint
        :return: whether the registration was a success or not
        :rtype: bool
        """
        pass

    @classmethod
    def unregister_candidate(cls, pubkey: ECPoint) -> bool:
        """
        Unregisters as a candidate.

        :param pubkey: The public key of the account to be unregistered
        :type pubkey: ECPoint
        :return: whether the unregistration was a success or not
        :rtype: bool
        """
        pass

    @classmethod
    def vote(cls, account: UInt160, vote_to: ECPoint) -> bool:
        """
        Votes for a candidate.

        :param account: the account that is voting
        :type account: UInt160
        :param vote_to: the public key of the one being voted
        :type vote_to: ECPoint
        """
        pass

    @classmethod
    def get_all_candidates(cls) -> Iterator:
        """
        Gets the registered candidates iterator.

        :return: all registered candidates
        :rtype: Iterator
        """
        pass

    @classmethod
    def un_vote(cls, account: UInt160) -> bool:
        """
        Removes the vote of the candidate voted. It would be the same as calling vote(account, None).

        :param account: the account that is removing the vote
        :type account: UInt160
        """
        pass

    @classmethod
    def get_candidates(cls) -> List[Tuple[ECPoint, int]]:
        """
        Gets the list of all registered candidates.

        :return: all registered candidates
        :rtype: List[Tuple[ECPoint, int]]
        """
        pass

    @classmethod
    def get_candidate_vote(cls, pubkey: ECPoint) -> int:
        """
        Gets votes from specific candidate.

        :return: Votes or -1 if it was not found.
        :rtype: int
        """
        pass

    @classmethod
    def get_committee(cls) -> List[ECPoint]:
        """
        Gets all committee members list.

        :return: all committee members
        :rtype: List[ECPoint]
        """
        pass

    @classmethod
    def get_next_block_validators(cls) -> List[ECPoint]:
        """
        Gets validators list of the next block.

        :return: the public keys of the validators
        :rtype: List[ECPoint]
        """
        pass

    @classmethod
    def get_account_state(cls, account: UInt160) -> NeoAccountState:
        """
        Gets the latest votes of the specified account.

        :param account: the specified account
        :type account: UInt160
        :return: the state of the account
        :rtype: NeoAccountState
        """
        pass



from typing import List, Tuple

from boa3.model.operation.binary.binaryoperation import BinaryOperation
from boa3.model.operation.operator import Operator
from boa3.model.type.type import IType, Type
from boa3.neo.vm.opcode.Opcode import Opcode


class ObjectInequality(BinaryOperation):
    """
    A class used to represent a inequality comparison

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: List[IType] = [Type.any]

    def __init__(self, left: IType = Type.str, right: IType = None):
        self.operator: Operator = Operator.NotEq
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return self._is_valid_type(left) and self._is_valid_type(right)

    def _is_valid_type(self, tpe: IType) -> bool:
        return any(valid.is_type_of(tpe) for valid in self._valid_types)

    def _get_result(self, left: IType, right: IType) -> IType:
        if self.validate_type(left, right):
            return Type.bool
        else:
            return Type.none

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        return [(Opcode.NOTEQUAL, b'')]
